(function(f){if(typeof exports==="object"&&typeof module!=="undefined"){module.exports=f()}else if(typeof define==="function"&&define.amd){define([],f)}else{var g;if(typeof window!=="undefined"){g=window}else if(typeof global!=="undefined"){g=global}else if(typeof self!=="undefined"){g=self}else{g=this}g.ecsLib = f()}})(function(){var define,module,exports;return (function(){function r(e,n,t){function o(i,f){if(!n[i]){if(!e[i]){var c="function"==typeof require&&require;if(!f&&c)return c(i,!0);if(u)return u(i,!0);var a=new Error("Cannot find module '"+i+"'");throw a.code="MODULE_NOT_FOUND",a}var p=n[i]={exports:{}};e[i][0].call(p.exports,function(r){var n=e[i][1][r];return o(n||r)},p,p.exports,r,e,n,t)}return n[i].exports}for(var u="function"==typeof require&&require,i=0;i<t.length;i++)o(t[i]);return o}return r})()({1:[function(require,module,exports){
"use strict";
var __extends = (this && this.__extends) || (function () {
    var extendStatics = function (d, b) {
        extendStatics = Object.setPrototypeOf ||
            ({ __proto__: [] } instanceof Array && function (d, b) { d.__proto__ = b; }) ||
            function (d, b) { for (var p in b) if (b.hasOwnProperty(p)) d[p] = b[p]; };
        return extendStatics(d, b);
    };
    return function (d, b) {
        extendStatics(d, b);
        function __() { this.constructor = d; }
        d.prototype = b === null ? Object.create(b) : (__.prototype = b.prototype, new __());
    };
})();
Object.defineProperty(exports, "__esModule", { value: true });
var NOW = function () { return 0; };
if (typeof window !== 'undefined' && window.performance) {
    NOW = window.performance.now.bind(window.performance);
}
else {
    var start_1 = Date.now();
    NOW = function () {
        return Date.now() - start_1;
    };
}
var SEQ_ENTITY = 1;
var SEQ_COMPONENT = 1;
/**
 * Representação de uma entidade o ECS
 */
var Entity = /** @class */ (function () {
    function Entity() {
        /**
         * Lista de interessados sobre a atualiação dos componentes
         */
        this.subscriptions = [];
        /**
         * Informa se a entidade está ativa
         */
        this.active = true;
        /**
         * Componentes por tipo
         *
         * ATENÇÃO! Apesar de a lista de componentes ser publica, não fazer modificação na mesma, utilizar os metodos add e
         * remove para tal
         */
        this.components = {};
        this.id = Entity.sequence();
    }
    Entity.sequence = function () {
        return SEQ_ENTITY++;
    };
    /**
     * Permite aos interessados receber informação quando a lista de componentes dessa entidade receber atualização
     *
     * @param handler
     */
    Entity.prototype.subscribe = function (handler) {
        var _this = this;
        this.subscriptions.push(handler);
        return function () {
            var idx = _this.subscriptions.indexOf(handler);
            if (idx >= 0) {
                _this.subscriptions.splice(idx, 1);
            }
        };
    };
    /**
     * Adiciona um componente à entidade
     *
     * @param component
     */
    Entity.prototype.add = function (component) {
        var _this = this;
        var type = component.type;
        if (!this.components[type]) {
            this.components[type] = [];
        }
        this.components[type].push(component);
        // Informa aos interessados sobre a atualização
        requestAnimationFrame(function () {
            _this.subscriptions.forEach(function (cb) { return cb(); });
        });
    };
    /**
     * Remove a referencia de um componente dessa entidade
     *
     * @param component
     */
    Entity.prototype.remove = function (component) {
        var _this = this;
        var type = component.type;
        if (!this.components[type]) {
            return;
        }
        var idx = this.components[type].indexOf(component);
        if (idx >= 0) {
            this.components[type].splice(idx, 1);
        }
        // Informa aos interessados sobre a atualização
        requestAnimationFrame(function () {
            _this.subscriptions.forEach(function (cb) { return cb(); });
        });
    };
    return Entity;
}());
exports.Entity = Entity;
/**
 * Informações sobre um componente
 */
var Component = /** @class */ (function () {
    function Component(type, data) {
        this.type = type;
        this.data = data;
    }
    /**
     * Registra uma nova classe de componente
     */
    Component.register = function () {
        var typeID = SEQ_COMPONENT++;
        var ComponentImpl = /** @class */ (function (_super) {
            __extends(ComponentImpl, _super);
            function ComponentImpl(data) {
                return _super.call(this, typeID, data) || this;
            }
            /**
             * Get all instances of this component from entity
             *
             * @param entity
             */
            ComponentImpl.allFrom = function (entity) {
                var components = entity.components[typeID];
                return components || [];
            };
            /**
             * Get one instance of this component from entity
             *
             * @param entity
             */
            ComponentImpl.oneFrom = function (entity) {
                var components = ComponentImpl.allFrom(entity);
                if (components && components.length > 0) {
                    return components[0];
                }
            };
            /**
             * Static reference to type
             */
            ComponentImpl.type = typeID;
            return ComponentImpl;
        }(Component));
        return ComponentImpl;
    };
    return Component;
}());
exports.Component = Component;
/**
 * Representação de um sistema no ECS
 */
var System = /** @class */ (function () {
    function System() {
        /**
         * Implementação deve informar quais tipo de componentes esse sistema opera sobre
         */
        this.components = [];
    }
    return System;
}());
exports.System = System;
/**
 * A propria definição do ECS. Também chamada de Admin ou Manager em outras implementações
 */
var ECS = /** @class */ (function () {
    function ECS(systems) {
        var _this = this;
        this.lastUpdate = NOW();
        /**
         * Todos os Systems existentes nesse mundo
         */
        this.systems = [];
        /**
         * Todas as entidades existentes neste mundo
         */
        this.entities = [];
        /**
         * Indexa os sistemas que devem ser executados para cada entidade
         */
        this.entitiesSystems = {};
        /**
         * Guarda as subscrições realizadas para entidades
         */
        this.entitiesSubscriptions = {};
        if (systems) {
            systems.forEach(function (system) {
                _this.addSystem(system);
            });
        }
    }
    /**
     * Obtém uma entidade por id
     *
     * @param id
     */
    ECS.prototype.getEntity = function (id) {
        return this.entities.find(function (entity) { return entity.id === id; });
    };
    /**
     * Adiciona uma entidade nesse mundo
     *
     * @param entity
     */
    ECS.prototype.addEntity = function (entity) {
        var _this = this;
        if (!entity) {
            return;
        }
        if (this.entities.indexOf(entity) < 0) {
            this.entities.push(entity);
            // Remove a subscrição, se existir
            if (this.entitiesSubscriptions[entity.id]) {
                this.entitiesSubscriptions[entity.id]();
            }
            // Adiciona nova subscrição
            this.entitiesSubscriptions[entity.id] = entity.subscribe(function () {
                _this.indexEntity(entity);
            });
        }
        this.indexEntity(entity);
    };
    /**
     * Remove uma entidade desse mundo
     *
     * @param idOrInstance
     */
    ECS.prototype.removeEntity = function (idOrInstance) {
        var entity = idOrInstance;
        if (typeof idOrInstance === 'number') {
            entity = this.getEntity(idOrInstance);
        }
        if (!entity) {
            return;
        }
        var idx = this.entities.indexOf(entity);
        if (idx >= 0) {
            this.entities.splice(idx, 1);
        }
        // Remove a subscrição, se existir
        if (this.entitiesSubscriptions[entity.id]) {
            this.entitiesSubscriptions[entity.id]();
        }
        // Invoca "exit" dos sistemas
        var systems = this.entitiesSystems[entity.id];
        if (systems) {
            systems.forEach(function (system) {
                if (system.exit) {
                    system.exit(entity);
                }
            });
        }
        // Remove índices associativos
        this.entitiesSystems[entity.id] = [];
    };
    /**
     * Adiciona o sistema nesse mundo
     *
     * @param system
     */
    ECS.prototype.addSystem = function (system) {
        var _this = this;
        if (!system) {
            return;
        }
        if (this.systems.indexOf(system) >= 0) {
            return;
        }
        this.systems.push(system);
        // Reindexa entidades
        this.entities.forEach(function (entity) {
            _this.indexEntity(entity, system);
        });
        // Invoca "enter" do sistema
        this.entities.forEach(function (entity) {
            if (!entity.active) {
                // Entidade inativa
                return _this.removeEntity(entity);
            }
            var systems = _this.entitiesSystems[entity.id];
            if (systems && systems.indexOf(system) >= 0) {
                if (system.enter) {
                    system.enter(entity);
                }
            }
        });
    };
    /**
     * Remove o sistema desse mundo
     *
     * @param system
     */
    ECS.prototype.removeSystem = function (system) {
        var _this = this;
        if (!system) {
            return;
        }
        var idx = this.systems.indexOf(system);
        if (idx >= 0) {
            // Invoca "exit" do sistema
            this.entities.forEach(function (entity) {
                if (!entity.active) {
                    // Entidade inativa
                    return _this.removeEntity(entity);
                }
                var systems = _this.entitiesSystems[entity.id];
                if (systems && systems.indexOf(system) >= 0) {
                    if (system.exit) {
                        system.exit(entity);
                    }
                }
            });
            this.systems.splice(idx, 1);
            // Reindexa entidades
            this.entities.forEach(function (entity) {
                _this.indexEntity(entity, system);
            });
        }
    };
    /**
     * Realiza a atualização dos sistemas desse mundo
     */
    ECS.prototype.update = function () {
        var _this = this;
        var now = NOW();
        var elapsed = now - this.lastUpdate;
        this.entities.forEach(function (entity) {
            if (!entity.active) {
                // Entidade inativa
                return _this.removeEntity(entity);
            }
            var systems = _this.entitiesSystems[entity.id];
            if (!systems) {
                return;
            }
            systems.forEach(function (system) {
                var components = {};
                system.components.forEach(function (typeId) {
                    components[typeId] = entity.components[typeId];
                });
                if (system.update) {
                    system.update(now, elapsed, entity);
                }
            });
        });
        this.lastUpdate = now;
    };
    /**
     * Faz a indexação de uma entidade
     *
     * @param entity
     */
    ECS.prototype.indexEntity = function (entity, system) {
        var _this = this;
        if (!this.entitiesSystems[entity.id]) {
            this.entitiesSystems[entity.id] = [];
        }
        var entityComponents = Object.keys(entity.components).map(function (v) { return Number.parseInt(v, 10); });
        var indexSystem = function (system) {
            var idx = _this.entitiesSystems[entity.id].indexOf(system);
            // Sistema não existe neste mundo, remove indexação
            if (_this.systems.indexOf(system) < 0) {
                if (idx >= 0) {
                    _this.entitiesSystems[entity.id].splice(idx, 1);
                }
                return;
            }
            var systemComponents = system.components;
            for (var a = 0, l = systemComponents.length; a < l; a++) {
                if (entityComponents.indexOf(systemComponents[a]) < 0) {
                    // remove
                    if (idx >= 0) {
                        // Informa ao sistema sobre a remoção do relacionamento
                        if (system.exit) {
                            system.exit(entity);
                        }
                        _this.entitiesSystems[entity.id].splice(idx, 1);
                    }
                    return;
                }
            }
            // Entidade possui todos os componentes que esse sistema precisa
            if (idx < 0) {
                _this.entitiesSystems[entity.id].push(system);
                // Informa ao sistema sobre o novo relacionamento
                if (system.enter) {
                    system.enter(entity);
                }
            }
        };
        if (system) {
            // Indexa a entidade para um sistema específico
            indexSystem(system);
        }
        else {
            // Reindexa toda a entidade
            this.systems.forEach(indexSystem);
        }
    };
    ECS.Entity = Entity;
    ECS.Component = Component;
    ECS.System = System;
    return ECS;
}());
exports.default = ECS;

},{}]},{},[1])(1)
});
